global phi
global psi

%% create training dataset, classification in R^d
K = 100;    % number of samples
d = 2;      % input dimension
%[x,t] = create_data(d,K);
            % x = point locations
            % t = class labels (targets)
figure(1); clf;
show_data(x,t);
            
%% initialize network (fully connected)
L = 3;      % depth
N = 5;      % width
net = cell(1,3);
net{1} = randn(d,N);    % alpha parameters (input layer)
net{2} = randn(N,N,L);  % w parameters (hidden layers)
net{3} = randn(1,N);    % beta parameters (output layer)
phi = @(z)1./(1+exp(-z));
psi = @(z)phi(z).*(1-phi(z));

%% train network

method = 'gdls';
        
switch method
    case 'gd'       % gradient descent
        tic
        nit = 2.5e4;
        res = zeros(1,nit);
        eta = 5e-3;     % learning rate
        for it = 1:nit;
            [y,z,u] = run_net(net, x);
            res(it) = sum((y-t).^2);
            dE = get_grad(net, y, z, u, x, t);
            for s = 1:3
                net{s} = net{s} - eta*dE{s};
            end
        end
        toc
    case 'sgd'      % stochastic gradient descent
        tic
        nit = 4e4;
        res = zeros(1,nit);
        eta = 1e-2;    % learning rate
        Q = K/5;     % minibatch size
        for it = 1:nit;    
            temp = randperm(K);
            Ibatch = temp(1:Q);     % with repetitions
            xbatch = x(:,Ibatch);
            tbatch = t(:,Ibatch);
            [y,z,u] = run_net(net, xbatch);
            res(it) = sum((y-tbatch).^2);      
            dE = get_grad(net, y, z, u, xbatch, tbatch);
           for s = 1:3
               net{s} = net{s} - eta*dE{s};
           end
        end
        toc
    case 'gdls'       % gradient descent, line search
        tic
        nit = 2e4;
        res = zeros(1,nit);
        etas = zeros(1,nit);        
        eta = 1;     % intial learning rate
        c1 = 1e-1;      % Wolfe's parameter of sufficient decrease
        prevnormdE2 = 1e-1;
        for it = 1:nit;
            [y,z,u] = run_net(net, x);
            res(it) = sum((y-t).^2);
            prevres = res(it);
            dE = get_grad(net, y, z, u, x, t);
            % line search initialization
            normdE2 = 0;
            for s = 1:3
                temp = reshape(dE{s},1,numel(dE{s}));
                normdE2 = normdE2 + sum(temp.^2);
            end
            eta = eta * sqrt(prevnormdE2/normdE2);
            prevnormdE2 = normdE2;
            % initial update
            newnet = cell(1,3);
            for s = 1:3
                newnet{s} = net{s} - eta*dE{s};
            end
            % line search
            wolfe = 0;
            while ~wolfe
                [y,z,u] = run_net(newnet, x);
                res(it) = sum((y-t).^2);
                wolfe = (res(it) - prevres < (-c1)*eta*normdE2);                 
                if ~wolfe, 
                    eta = eta/2;
                    newnet = cell(1,3);
                    for s = 1:3
                        newnet{s} = net{s} - eta*dE{s};
                    end
                end                
            end            
            net = newnet;
            etas(it) = eta;
        end
        toc

    case 'bfgs'
    case 'nlcg'
end

figure(3); clf;
loglog(1:nit, res,'linewidth',3);

if exist('etas')
    figure(4); clf;
    semilogx(1:nit, etas,'linewidth',3);
end
clear etas

%% test the network
h = 1e-2;   % grid spacing
xfine = 0:h:1;
[xx1, xx2] = meshgrid(xfine,xfine);       % assumes d = 2
x1 = reshape(xx1,1,numel(xx1));
x2 = reshape(xx2,1,numel(xx1));
xtest = [x1;x2];
ytest = run_net(net, xtest);
yy = reshape(ytest,length(xfine),length(xfine));

%figure(1); hold on;
%mesh(xx1,xx2,yy);

figure(2); clf;
imagesc(xfine,xfine,yy);
set(gca,'Ydir','normal')
colormap(1-gray)
colorbar


